# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging
import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

import gi
from gi.repository import GLib
from gi.repository import GObject

from gajim.common.enum import AudioPlayerState
from gajim.common.ged import EventHelper

try:
    gi.require_version("Gst", "1.0")
    from gi.repository import Gst
except Exception:
    if typing.TYPE_CHECKING:
        from gi.repository import Gst

from gajim.gtk.util.classes import SignalManager

AudioSampleT = list[tuple[float, float]]

SEEK_BAR_PADDING = 11

log = logging.getLogger("gajim.gtk.preview.audio_player")


@dataclass
class AudioPreviewState:
    duration: float = 0.0
    position: float = 0.0
    pipeline_state: AudioPlayerState = AudioPlayerState.NULL
    speed: float = 1.0
    is_timestamp_positive: bool = True
    samples: AudioSampleT = field(default_factory=list[tuple[float, float]])
    is_audio_analyzed = False


class AudioPlayer(GObject.GObject, SignalManager, EventHelper):
    __gtype_name__ = "AudioPlayer"
    __gsignals__ = {
        "audio-playback-changed": (GObject.SignalFlags.RUN_LAST, None, (int, int)),
        "audio-playback-progressed": (GObject.SignalFlags.RUN_LAST, None, (int, int)),
    }

    def __init__(self) -> None:
        super().__init__()
        SignalManager.__init__(self)
        EventHelper.__init__(self)

        self._file_path = None
        self._state = None
        self._preview_id = -1

        self._speed_min = 0.25
        self._speed_max = 2.00

        self._playbin = Gst.ElementFactory.make("playbin", "bin")
        self._query = Gst.Query.new_position(Gst.Format.TIME)
        self._bus_watch_id: int = 0

        self._do_broadcasting = False
        self._switch_track = False
        self._play_new_track = False
        self._seek_track_position = False
        self._setup_pipeline()

        # Holds active audio preview sessions
        # for resuming after switching chats
        self._audio_sessions: dict[int, AudioPreviewState] = {}

    def destroy(self) -> None:
        if self._playbin is not None:
            self._playbin.set_state(Gst.State.NULL)
            bus = self._playbin.get_bus()

            if bus is not None:
                bus.remove_signal_watch()
                bus.disconnect(self._bus_watch_id)

    @property
    def preview_id(self) -> int:
        return self._preview_id

    def get_audio_state(self, preview_id: int) -> AudioPreviewState:
        state = self._audio_sessions.get(preview_id)
        if state is not None:
            return state
        self._audio_sessions[preview_id] = AudioPreviewState()
        return self._audio_sessions[preview_id]

    def toggle_playback(self) -> None:
        assert self._playbin is not None

        if self._is_playing():
            self._playbin.set_state(Gst.State.PAUSED)
        else:
            if self._is_eos():
                # Treat re-playing same as playing a new track
                self._switch_track = True
                self._seek_track_position = True
                self._play_new_track = True
                self._set_playback_position(0.0)
            else:
                self._playbin.set_state(Gst.State.PLAYING)

    def play_audio_file(self, file_path: Path, preview_id: int) -> None:
        assert self._playbin is not None
        assert preview_id != self._preview_id

        self._do_broadcasting = False
        self._playbin.set_state(Gst.State.READY)

        self._preview_id = preview_id

        state = self._audio_sessions.get(preview_id)
        assert state is not None
        self._state = state
        self._file_path = file_path

        self._switch_track = True
        self._seek_track_position = True
        self._play_new_track = True

        self._playbin.set_property("uri", self._file_path.as_uri())

        if self._is_eos():
            self._state.position = 0.0
        self._playbin.set_state(Gst.State.PLAYING)

    def pause(self) -> None:
        assert self._playbin is not None
        self._playbin.set_state(Gst.State.PAUSED)

    def stop(self, preview_id: int) -> None:
        assert self._playbin is not None

        if self._preview_id != preview_id:
            return

        self.stop_any()

    def stop_any(self):
        assert self._playbin is not None

        self._do_broadcasting = False
        self._preview_id = -1
        self._playbin.set_state(Gst.State.NULL)
        if self._state is not None:
            self._state.pipeline_state = AudioPlayerState.NULL
        self.emit(
            "audio-playback-changed",
            -1,
            AudioPlayerState.NULL.value,
        )

    def set_playback_position(self, preview_id: int, position: float) -> None:
        assert self._playbin is not None

        if preview_id != self._preview_id or self._state is None:
            return

        self._set_playback_position(position)

    def set_playback_speed(self, speed: float) -> bool:
        assert self._state is not None

        if self._speed_min <= speed <= self._speed_max:
            self._state.speed = speed

        assert self._playbin is not None
        self._playbin.seek(
            self._state.speed,
            Gst.Format.TIME,
            Gst.SeekFlags.FLUSH,
            Gst.SeekType.SET,
            int(self._state.position),
            Gst.SeekType.NONE,
            0,
        )
        return True

    def _is_eos(self) -> bool:
        assert self._state is not None
        return abs(self._state.position - self._state.duration) < 0.1e9

    def _query_position(self) -> float:
        assert self._playbin is not None
        assert self._state is not None

        if self._playbin.query(self._query) and (
            self._is_paused() or self._is_playing()
        ):
            _fmt, position = self._query.parse_position()
            return position

        return self._state.position

    def _map_gst_state_to_audio_state(self, gst_state: Gst.State) -> AudioPlayerState:
        if gst_state == Gst.State.NULL:
            return AudioPlayerState.NULL
        elif gst_state == Gst.State.READY:
            return AudioPlayerState.READY
        elif gst_state == Gst.State.PAUSED:
            return AudioPlayerState.PAUSED
        else:
            return AudioPlayerState.PLAYING

    def _setup_pipeline(self) -> None:
        assert self._playbin is not None
        # Set up the whole pipeline
        # For reference see
        # https://gstreamer.freedesktop.org/
        # documentation/audiofx/scaletempo.html
        audio_sink = Gst.Bin.new("audiosink")
        audioconvert = Gst.ElementFactory.make("audioconvert", "audioconvert")
        scaletempo = Gst.ElementFactory.make("scaletempo", "scaletempo")
        audioresample = Gst.ElementFactory.make("audioresample", "audioresample")
        autoaudiosink = Gst.ElementFactory.make("autoaudiosink", "autoaudiosink")

        pipeline_elements = [
            audio_sink,
            audioconvert,
            scaletempo,
            audioresample,
            autoaudiosink,
        ]

        if any(element is None for element in pipeline_elements):
            # If it fails there will be
            # * a delay until playback starts
            # * a chipmunk effect when speeding up the playback
            log.error("Could not set up full audio preview pipeline.")
        else:
            assert autoaudiosink is not None
            assert audioconvert is not None
            assert scaletempo is not None
            assert audioresample is not None

            # On Windows the first fraction of an audio
            # would not play if not synced.
            # On Linux there can be a delay before playback starts with sync,
            # which is however not the case on Windows.
            if sys.platform == "win32":
                autoaudiosink.set_property("sync", True)
            else:
                autoaudiosink.set_property("sync", False)

            audio_sink.add(audioconvert)
            audio_sink.add(scaletempo)
            audio_sink.add(audioresample)
            audio_sink.add(autoaudiosink)

            audioconvert.link(scaletempo)
            scaletempo.link(audioresample)
            audioresample.link(autoaudiosink)

            sink_pad = audioconvert.get_static_pad("sink")
            assert sink_pad is not None
            ghost_pad = Gst.GhostPad.new("sink", sink_pad)
            assert ghost_pad is not None
            audio_sink.add_pad(ghost_pad)
            self._playbin.set_property("audio-sink", audio_sink)

        self._playbin.no_more_pads()

        state_return = self._playbin.set_state(Gst.State.READY)
        if state_return == Gst.StateChangeReturn.FAILURE:
            log.error("Could not setup GST playbin.")
            return

        bus = self._playbin.get_bus()
        if bus is None:
            log.error("Could not get GST Bus.")
            return

        bus.add_signal_watch()
        self._bus_watch_id = bus.connect("message", self._on_bus_message)
        self._pipeline_is_setup = True

    def _is_paused(self) -> bool:
        assert self._playbin is not None
        _, state, _ = self._playbin.get_state(timeout=40)
        return state == Gst.State.PAUSED

    def _is_playing(self) -> bool:
        assert self._playbin is not None
        _, state, _ = self._playbin.get_state(timeout=40)
        return state == Gst.State.PLAYING

    def _clip_position(self, pos: float) -> float:
        assert self._state is not None
        if pos >= self._state.duration:
            return self._state.duration
        if pos < 0:
            return 0.0
        return pos

    def _set_playback_position(self, position: float) -> None:
        assert self._state is not None
        assert self._playbin is not None

        self._state.position = self._clip_position(position)
        self._query.set_position(Gst.Format.TIME, int(self._state.position))
        self._playbin.seek(
            self._state.speed,
            Gst.Format.TIME,
            Gst.SeekFlags.FLUSH,
            Gst.SeekType.SET,
            int(self._state.position),
            Gst.SeekType.NONE,
            0,
        )

    def _broadcast_playback_progression(self):
        if not self._do_broadcasting:
            return False

        assert self._state is not None
        self._state.position = self._query_position()
        self.emit(
            "audio-playback-progressed",
            self._preview_id,
            self._state.pipeline_state.value,
        )
        return True

    def _on_bus_message(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        assert self._playbin is not None

        if (
            message.type == Gst.MessageType.ASYNC_DONE
            and self._file_path is not None
            and self._state is not None
        ):
            if self._switch_track:
                self._playbin.set_state(Gst.State.READY)
                self._playbin.set_property("uri", self._file_path.as_uri())
                self._query.set_position(Gst.Format.TIME, int(self._state.position))
                self._switch_track = False
                self._playbin.set_state(Gst.State.PAUSED)
            elif self._seek_track_position:
                self._set_playback_position(self._state.position)
                self._seek_track_position = False
            elif self._play_new_track:
                self._play_new_track = False
                self._playbin.set_state(Gst.State.PLAYING)

        if message.type == Gst.MessageType.STATE_CHANGED:
            old_state, new_state, _pending_state = message.parse_state_changed()
            if message.src == self._playbin and old_state != new_state:
                if self._state is None:
                    return

                self._state.pipeline_state = self._map_gst_state_to_audio_state(
                    new_state
                )
                self.emit(
                    "audio-playback-changed",
                    self._preview_id,
                    self._state.pipeline_state.value,
                )
                log.debug(
                    "State changed: %s → %s",
                    Gst.Element.state_get_name(old_state),
                    Gst.Element.state_get_name(new_state),
                )

                if new_state == Gst.State.PLAYING:
                    self._do_broadcasting = True
                    GLib.timeout_add(20, self._broadcast_playback_progression)
                    self.emit(
                        "audio-playback-changed",
                        self._preview_id,
                        self._state.pipeline_state.value,
                    )
                    return
                self._do_broadcasting = False
            return

        if message.type == Gst.MessageType.EOS and self._state is not None:
            self._state.position = self._state.duration
            self._playbin.set_state(Gst.State.PAUSED)
            self._do_broadcasting = False
            return

        if message.type == Gst.MessageType.ERROR:
            self._handle_error_on_playback(message)

    def _error_info(self, message: Gst.Message) -> tuple[str, int]:
        structure = message.get_structure()
        assert structure is not None
        gerror = structure.get_value("gerror")
        assert gerror is not None
        return gerror.domain, gerror.code

    def _handle_error_on_playback(self, message: Gst.Message):
        assert self._playbin is not None
        domain, code = self._error_info(message)
        if domain == "gst-resource-error-quark":
            if code == 10:
                # On Windows: AUDCLNT_E_DEVICE_INVALIDATED (10)
                # GST_RESOURCE_ERROR_WRITE (10)
                # used when the resource can't be written to.
                self._playbin.set_state(Gst.State.READY)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Types for typechecking

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeVar

import cairo

if TYPE_CHECKING:
    from gajim.gtk.control import ChatControl
    from gajim.gtk.conversation.message_widget import MessageWidget
    from gajim.gtk.conversation.rows.call import CallRow
    from gajim.gtk.conversation.rows.command_output import CommandOutputRow
    from gajim.gtk.conversation.rows.date import DateRow
    from gajim.gtk.conversation.rows.file_transfer import FileTransferRow
    from gajim.gtk.conversation.rows.file_transfer_jingle import FileTransferJingleRow
    from gajim.gtk.conversation.rows.info import InfoMessage
    from gajim.gtk.conversation.rows.message import MessageRow
    from gajim.gtk.conversation.rows.muc_join_left import MUCJoinLeft
    from gajim.gtk.conversation.rows.muc_subject import MUCSubject
    from gajim.gtk.conversation.rows.scroll_hint import ScrollHintRow
    from gajim.gtk.conversation.rows.user_status import UserStatus
    from gajim.gtk.conversation.view import ConversationView


SomeSurface = TypeVar("SomeSurface", bound=cairo.Surface)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import overload

import logging
from functools import partial
from unittest.mock import Mock

from nbxmpp.dispatcher import NBXMPPModuleNameT
from nbxmpp.dispatcher import NBXMPPModuleT
from nbxmpp.modules.activity import Activity
from nbxmpp.modules.adhoc import AdHoc
from nbxmpp.modules.annotations import Annotations
from nbxmpp.modules.attention import Attention
from nbxmpp.modules.base import BaseModule as NBXMPPBaseModule
from nbxmpp.modules.blocking import Blocking
from nbxmpp.modules.bookmarks.native_bookmarks import NativeBookmarks
from nbxmpp.modules.bookmarks.pep_bookmarks import PEPBookmarks
from nbxmpp.modules.bookmarks.private_bookmarks import PrivateBookmarks
from nbxmpp.modules.captcha import Captcha
from nbxmpp.modules.chat_markers import ChatMarkers
from nbxmpp.modules.chatstates import Chatstates
from nbxmpp.modules.correction import Correction
from nbxmpp.modules.delay import Delay
from nbxmpp.modules.delimiter import Delimiter
from nbxmpp.modules.discovery import Discovery
from nbxmpp.modules.eme import EME
from nbxmpp.modules.entity_caps import EntityCaps
from nbxmpp.modules.entity_time import EntityTime
from nbxmpp.modules.http_auth import HTTPAuth
from nbxmpp.modules.http_upload import HTTPUpload
from nbxmpp.modules.ibb import IBB
from nbxmpp.modules.idle import Idle
from nbxmpp.modules.iq import BaseIq
from nbxmpp.modules.last_activity import LastActivity
from nbxmpp.modules.location import Location
from nbxmpp.modules.mam import MAM
from nbxmpp.modules.mds import MDS
from nbxmpp.modules.message import BaseMessage
from nbxmpp.modules.mood import Mood
from nbxmpp.modules.muc.hats import Hats
from nbxmpp.modules.muc.moderation import Moderation
from nbxmpp.modules.muc.muc import MUC
from nbxmpp.modules.muclumbus import Muclumbus
from nbxmpp.modules.nickname import Nickname
from nbxmpp.modules.omemo import OMEMO
from nbxmpp.modules.oob import OOB
from nbxmpp.modules.openpgp import OpenPGP
from nbxmpp.modules.pgplegacy import PGPLegacy
from nbxmpp.modules.ping import Ping
from nbxmpp.modules.presence import BasePresence
from nbxmpp.modules.pubsub import PubSub
from nbxmpp.modules.reactions import Reactions
from nbxmpp.modules.receipts import Receipts
from nbxmpp.modules.register.register import Register
from nbxmpp.modules.replies import Replies
from nbxmpp.modules.retraction import Retraction
from nbxmpp.modules.roster import Roster
from nbxmpp.modules.security_labels import SecurityLabels
from nbxmpp.modules.software_version import SoftwareVersion
from nbxmpp.modules.tune import Tune
from nbxmpp.modules.user_avatar import UserAvatar
from nbxmpp.modules.vcard4 import VCard4
from nbxmpp.modules.vcard_avatar import VCardAvatar
from nbxmpp.modules.vcard_temp import VCardTemp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.ged import EventHelper
from gajim.common.modules.util import LogAdapter


class BaseModule(EventHelper):

    _nbxmpp_extends: NBXMPPModuleNameT = ''  # type: ignore
    _nbxmpp_methods: list[str] = []

    def __init__(self,
                 con: types.Client,
                 *args: Any,
                 plugin: bool = False,
                 **kwargs: Any) -> None:

        EventHelper.__init__(self)
        self._con = con
        self._client = con
        self._account = con.account
        self._log = self._set_logger(plugin)
        self._nbxmpp_callbacks: dict[str, Any] = {}
        self._stored_publish: types.AnyCallableT | None = None
        self._ttl_cache: dict[Any, Any] = {}
        self.handlers: list[StanzaHandler] = []

    @classmethod
    def get_instance(cls, client: types.Client) -> BaseModule:
        return cls(client)

    def get_ttl_cache(self) -> dict[Any, Any]:
        return self._ttl_cache

    def _set_logger(self, plugin: bool) -> LogAdapter:
        logger_name = 'gajim.c.m.%s'
        if plugin:
            logger_name = 'gajim.p.%s'
        logger_name = logger_name % self.__class__.__name__.lower()
        logger = logging.getLogger(logger_name)
        return LogAdapter(logger, {'account': self._account})

    def __getattr__(self, key: str) -> Any:
        if key not in self._nbxmpp_methods:
            raise AttributeError(
                f'attribute "{key}" is neither part of object '
                f'"{self.__class__.__name__}" nor declared in '
                '"_nbxmpp_methods"')

        if not app.account_is_connected(self._account):
            self._log.warning('Account not connected, can’t use %s', key)
            return None

        module = cast(
            NBXMPPBaseModule,
            self._con.connection.get_module(self._nbxmpp_extends)
        )

        callback = self._nbxmpp_callbacks.get(key)
        if callback is None:
            return getattr(module, key)
        return partial(getattr(module, key), callback=callback)

    def _register_callback(self, method: str, callback: types.AnyCallableT) -> None:
        self._nbxmpp_callbacks[method] = callback

    def _register_pubsub_handler(self, callback: types.AnyCallableT) -> None:
        handler = StanzaHandler(name='message',
                                callback=callback,
                                ns=Namespace.PUBSUB_EVENT,
                                priority=49)
        self.handlers.append(handler)

    def send_stored_publish(self) -> None:
        if self._stored_publish is None:
            return
        self._log.info('Send stored publish')
        self._stored_publish()  # pylint: disable=not-callable

    def _get_contact(self,
                     jid: JID,
                     groupchat: bool = False) -> types.ContactT:
        return self._con.get_module('Contacts').get_contact(
            jid, groupchat=groupchat)

    def _get_contact_if_exists(self, jid: JID) -> types.ContactT | None:
        return self._con.get_module('Contacts').get_contact_if_exists(jid)

    def _get_own_bare_jid(self) -> JID:
        return self._client.get_own_jid().new_as_bare()

    def cleanup(self) -> None:
        self._ttl_cache.clear()
        self.handlers.clear()
        self.unregister_events()
        self._client.disconnect_all_from_obj(self)

    @overload
    def _nbxmpp(self, name: Literal['Activity']) -> Activity: ...
    @overload
    def _nbxmpp(self, name: Literal['AdHoc']) -> AdHoc: ...
    @overload
    def _nbxmpp(self, name: Literal['Annotations']) -> Annotations: ...
    @overload
    def _nbxmpp(self, name: Literal['Attention']) -> Attention: ...
    @overload
    def _nbxmpp(self, name: Literal['Blocking']) -> Blocking: ...
    @overload
    def _nbxmpp(self, name: Literal['NativeBookmarks'] ) -> NativeBookmarks: ...  # noqa: E501
    @overload
    def _nbxmpp(self, name: Literal['PEPBookmarks']) -> PEPBookmarks: ...
    @overload
    def _nbxmpp(self, name: Literal['PrivateBookmarks']) -> PrivateBookmarks: ...  # noqa: E501
    @overload
    def _nbxmpp(self, name: Literal['Captcha']) -> Captcha: ...
    @overload
    def _nbxmpp(self, name: Literal['ChatMarkers']) -> ChatMarkers: ...
    @overload
    def _nbxmpp(self, name: Literal['Chatstates']) -> Chatstates: ...
    @overload
    def _nbxmpp(self, name: Literal['Correction']) -> Correction: ...
    @overload
    def _nbxmpp(self, name: Literal['Delay']) -> Delay: ...
    @overload
    def _nbxmpp(self, name: Literal['Delimiter']) -> Delimiter: ...
    @overload
    def _nbxmpp(self, name: Literal['Discovery']) -> Discovery: ...
    @overload
    def _nbxmpp(self, name: Literal['EME']) -> EME: ...
    @overload
    def _nbxmpp(self, name: Literal['EntityCaps']) -> EntityCaps: ...
    @overload
    def _nbxmpp(self, name: Literal['EntityTime']) -> EntityTime: ...
    @overload
    def _nbxmpp(self, name: Literal['Hats']) -> Hats: ...
    @overload
    def _nbxmpp(self, name: Literal['HTTPAuth']) -> HTTPAuth: ...
    @overload
    def _nbxmpp(self, name: Literal['HTTPUpload']) -> HTTPUpload: ...
    @overload
    def _nbxmpp(self, name: Literal['IBB']) -> IBB: ...
    @overload
    def _nbxmpp(self, name: Literal['Idle']) -> Idle: ...
    @overload
    def _nbxmpp(self, name: Literal['BaseIq']) -> BaseIq: ...
    @overload
    def _nbxmpp(self, name: Literal['LastActivity']) -> LastActivity: ...
    @overload
    def _nbxmpp(self, name: Literal['Location']) -> Location: ...
    @overload
    def _nbxmpp(self, name: Literal['MAM']) -> MAM: ...
    @overload
    def _nbxmpp(self, name: Literal['BaseMessage']) -> BaseMessage: ...
    @overload
    def _nbxmpp(self, name: Literal['MDS']) -> MDS: ...
    @overload
    def _nbxmpp(self, name: Literal['Mood']) -> Mood: ...
    @overload
    def _nbxmpp(self, name: Literal['MUC']) -> MUC: ...
    @overload
    def _nbxmpp(self, name: Literal['Moderation']) ->Moderation : ...
    @overload
    def _nbxmpp(self, name: Literal['Muclumbus']) -> Muclumbus: ...
    @overload
    def _nbxmpp(self, name: Literal['Nickname']) -> Nickname: ...
    @overload
    def _nbxmpp(self, name: Literal['OMEMO']) -> OMEMO: ...
    @overload
    def _nbxmpp(self, name: Literal['OOB']) -> OOB: ...
    @overload
    def _nbxmpp(self, name: Literal['OpenPGP']) -> OpenPGP: ...
    @overload
    def _nbxmpp(self, name: Literal['PGPLegacy']) -> PGPLegacy: ...
    @overload
    def _nbxmpp(self, name: Literal['Ping']) -> Ping: ...
    @overload
    def _nbxmpp(self, name: Literal['BasePresence']) -> BasePresence: ...
    @overload
    def _nbxmpp(self, name: Literal['PubSub']) -> PubSub: ...
    @overload
    def _nbxmpp(self, name: Literal['Reactions']) -> Reactions: ...
    @overload
    def _nbxmpp(self, name: Literal['Receipts']) -> Receipts: ...
    @overload
    def _nbxmpp(self, name: Literal['Register']) -> Register: ...
    @overload
    def _nbxmpp(self, name: Literal['Replies']) -> Replies: ...
    @overload
    def _nbxmpp(self, name: Literal['Retraction']) -> Retraction: ...
    @overload
    def _nbxmpp(self, name: Literal['Roster']) -> Roster: ...
    @overload
    def _nbxmpp(self, name: Literal['SecurityLabels']) -> SecurityLabels: ...
    @overload
    def _nbxmpp(self, name: Literal['SoftwareVersion']) -> SoftwareVersion: ...
    @overload
    def _nbxmpp(self, name: Literal['Tune']) -> Tune: ...
    @overload
    def _nbxmpp(self, name: Literal['UserAvatar']) -> UserAvatar: ...
    @overload
    def _nbxmpp(self, name: Literal['VCard4']) -> VCard4: ...
    @overload
    def _nbxmpp(self, name: Literal['VCardAvatar']) -> VCardAvatar: ...
    @overload
    def _nbxmpp(self, name: Literal['VCardTemp']) -> VCardTemp: ...
    @overload
    def _nbxmpp(self) -> types.NBXMPPClient: ...

    def _nbxmpp(self,
                name: NBXMPPModuleNameT | None = None
                ) -> Mock | types.NBXMPPClient | NBXMPPModuleT:

        if not app.account_is_connected(self._client.account):
            self._log.warning('Account not connected, can’t use nbxmpp method')
            return Mock()

        if name is None:
            return self._client.connection

        return self._client.connection.get_module(name)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload

import logging

from gi.repository import Gtk
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import RegisterStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.protocol import JID
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantProgressPage
from gajim.gtk.assistant import AssistantSuccessPage
from gajim.gtk.dataform import DataFormWidget

log = logging.getLogger("gajim.gtk.service_registration")


class ServiceRegistration(Assistant):
    def __init__(self, account: str, address: JID) -> None:
        Assistant.__init__(self, width=600, height=400)
        self.account = account

        self._client = app.get_client(account)
        self._service = address

        self.add_button("back", _("Back"))
        self.add_button("close", _("Close"))
        self.add_button("register", _("Register"), "suggested-action")

        self.set_button_visible_func(self._visible_func)

        self.add_pages(
            {
                "form": Form(),
            }
        )

        self.add_default_page("error")
        self.add_default_page("progress")
        self.add_default_page("success")

        self.get_page("error").set_title(_("Registration failed"))
        self.get_page("error").set_heading(_("Registration failed"))

        self._connect(self, "button-clicked", self._on_button_clicked)

        self._request_form()

    @overload
    def get_page(self, name: Literal["form"]) -> Form: ...

    @overload
    def get_page(self, name: Literal["success"]) -> AssistantSuccessPage: ...

    @overload
    def get_page(self, name: Literal["error"]) -> AssistantErrorPage: ...

    @overload
    def get_page(self, name: Literal["progress"]) -> AssistantProgressPage: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    @staticmethod
    def _visible_func(_assistant: Assistant, page_name: str) -> list[str]:
        if page_name == "form":
            return ["close", "register"]

        if page_name == "progress":
            return []

        if page_name == "success":
            return ["close"]

        if page_name == "error":
            return ["back", "close"]
        raise ValueError(f"page {page_name} unknown")

    def _on_button_clicked(self, _page: Gtk.Widget, button_name: str) -> None:
        if button_name == "register":
            self._register()

        if button_name == "back":
            self.show_page("form", Gtk.StackTransitionType.SLIDE_RIGHT)

        if button_name == "close":
            self.close()

    def _request_form(self) -> None:
        self.get_page("progress").set_title(_("Requesting Register Form"))
        self.get_page("progress").set_text(_("Requesting register form from server…"))
        self.show_page("progress")

        self._client.get_module("Register").request_register_form(
            self._service, callback=self._on_register_form
        )

    def _on_register_form(self, task: Task) -> None:
        try:
            result = task.finish()
        except (StanzaError, MalformedStanzaError) as error:
            self.get_page("error").set_text(error.get_text())
            self.show_page("error")
            return

        form = result.form
        if result.form is None:
            form = result.fields_form

        self.get_page("form").add_form(form)
        self.show_page("form")

    def _register(self) -> None:
        self.get_page("progress").set_title(_("Registering…"))
        self.get_page("progress").set_text(_("Registering…"))
        self.show_page("progress")

        form = self.get_page("form").get_submit_form()
        self._client.get_module("Register").submit_register_form(
            form, self._service, callback=self._on_register_result
        )

    def _on_register_result(self, task: Task) -> None:
        try:
            task.finish()
        except (StanzaError, MalformedStanzaError, RegisterStanzaError) as error:
            self.get_page("error").set_text(error.get_text())
            self.show_page("error")
            return

        self.get_page("success").set_title(_("Registration successful"))
        self.get_page("success").set_text(_("Registration successful"))
        self.show_page("success")


class Form(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Register")
        self.complete = False

        self._dataform_widget: DataFormWidget | None = None

    def add_form(self, form: Any) -> None:
        self.remove_form()

        options = {"form-width": 350, "entry-activates-default": True}
        self._dataform_widget = DataFormWidget(form, options)
        self._dataform_widget.set_propagate_natural_height(True)
        self._connect(self._dataform_widget, "is-valid", self._on_is_valid)
        self._dataform_widget.validate()
        self.append(self._dataform_widget)

    def _on_is_valid(self, _widget: DataFormWidget, is_valid: bool) -> None:
        self.complete = is_valid
        self.update_page_complete()

    def get_submit_form(self) -> Any:
        assert self._dataform_widget is not None
        return self._dataform_widget.get_submit_form()

    def remove_form(self) -> None:
        if self._dataform_widget is None:
            return

        self.remove(self._dataform_widget)
        self._dataform_widget = None

    def get_default_button(self) -> str:
        return "register"



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import os
from dataclasses import dataclass

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.modes import GCM


@dataclass
class AESKeyData:
    key: bytes
    iv: bytes

    @classmethod
    def init(cls, key_size: int = 32, iv_size: int = 12) -> AESKeyData:
        key = os.urandom(key_size)
        iv = os.urandom(iv_size)
        return cls(key=key, iv=iv)


class AESGCMDecryptor:
    def __init__(self, data: AESKeyData) -> None:
        self._cipher = Cipher(
            algorithms.AES(data.key), GCM(data.iv), backend=default_backend()
        ).decryptor()
        self._cache = b""

    def decrypt(self, data: bytes) -> bytes:
        self._cache += data
        # Never use the last 16 bytes, because it could be the auth tag
        ciphertext = self._cache[:-16]
        self._cache = data[-16:]
        return self._cipher.update(ciphertext)

    def finalize(self) -> bytes:
        return self._cipher.finalize_with_tag(self._cache)


class AESGCMEncryptor:
    def __init__(self, data: AESKeyData) -> None:
        self._cipher = Cipher(
            algorithms.AES(data.key), GCM(data.iv), backend=default_backend()
        ).encryptor()

    def encrypt(self, data: bytes) -> bytes:
        return self._cipher.update(data)

    def finalize(self) -> bytes:
        return self._cipher.finalize() + self._cipher.tag



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

from enum import IntEnum
from pathlib import Path

from gi.repository import Gtk

from gajim.common import app
from gajim.common.helpers import check_soundfile_path
from gajim.common.helpers import play_sound
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.filechoosers import FileChooserButton
from gajim.gtk.filechoosers import Filter
from gajim.gtk.window import GajimAppWindow

SOUNDS = {
    "attention_received": _("Attention Message Received"),
    "first_message_received": _("Message Received"),
    "contact_connected": _("Contact Connected"),
    "contact_disconnected": _("Contact Disconnected"),
    "message_sent": _("Message Sent"),
    "muc_message_highlight": _("Group Chat Message Highlight"),
    "muc_message_received": _("Group Chat Message Received"),
    "incoming-call-sound": _("Call Incoming"),
    "outgoing-call-sound": _("Call Outgoing"),
}


class Column(IntEnum):
    ENABLED = 0
    NAME = 1
    PATH = 2
    CONFIG = 3


class ManageSounds(GajimAppWindow):
    def __init__(self, transient_for: Gtk.Window | None = None) -> None:
        GajimAppWindow.__init__(
            self,
            name="ManageSounds",
            title=_("Manage Sounds"),
            default_width=400,
            default_height=400,
            transient_for=transient_for,
            modal=True,
            add_window_padding=True,
            header_bar=True,
        )

        self._ui = get_builder("manage_sounds.ui")
        self.set_child(self._ui.manage_sounds)

        liststore = Gtk.ListStore(bool, str, str, str)
        self._ui.sounds_treeview.set_model(liststore)

        last_path = app.settings.get("last_sounds_dir")
        last_path = Path(last_path) if last_path else None

        self._file_chooser_button = FileChooserButton(
            filters=[
                Filter(name=_("All files"), patterns=["*"]),
                Filter(name=_("WAV Sounds"), suffixes=["wav"], default=True),
            ],
            label=_("Choose Sound"),
            initial_path=last_path,
        )
        self._file_chooser_button.set_hexpand(True)
        self._ui.sound_buttons_box.prepend(self._file_chooser_button)

        self._connect(liststore, "row-changed", self._on_row_changed)
        self._connect(
            self._ui.sounds_treeview, "cursor-changed", self._on_cursor_changed
        )
        self._connect(self._ui.toggle_cell_renderer, "toggled", self._on_toggle)
        self._connect(self._ui.clear_sound_button, "clicked", self._on_clear)
        self._connect(self._ui.play_sound_button, "clicked", self._on_play)
        self._connect(self._file_chooser_button, "path-picked", self._on_path_picked)

        self._fill_sound_treeview()

    @staticmethod
    def _on_row_changed(
        model: Gtk.TreeModel, path: Gtk.TreePath, iter_: Gtk.TreeIter
    ) -> None:
        sound_event = model[iter_][Column.CONFIG]
        app.settings.set_soundevent_setting(
            sound_event, "enabled", model[path][Column.ENABLED]
        )
        app.settings.set_soundevent_setting(
            sound_event, "path", model[iter_][Column.PATH]
        )

    def _on_toggle(self, _cell: Gtk.CellRendererToggle, path: Gtk.TreePath) -> None:
        if self._file_chooser_button.get_path() is None:
            return

        model = self._ui.sounds_treeview.get_model()
        assert model is not None

        model[path][Column.ENABLED] = not model[path][Column.ENABLED]

    def _fill_sound_treeview(self) -> None:
        model = cast(Gtk.ListStore, self._ui.sounds_treeview.get_model())
        model.clear()

        for sound_event, sound_name in SOUNDS.items():
            settings = app.settings.get_soundevent_settings(sound_event)
            model.append(
                [settings["enabled"], sound_name, settings["path"], sound_event]
            )

    def _on_cursor_changed(self, treeview: Gtk.TreeView) -> None:
        model, iter_ = treeview.get_selection().get_selected()
        assert iter_ is not None

        path_to_snd_file = check_soundfile_path(model[iter_][Column.PATH])
        if path_to_snd_file is None:
            self._file_chooser_button.reset()
            last_path = app.settings.get("last_sounds_dir")
            if last_path:
                self._file_chooser_button.set_inital_path(Path(last_path))
        else:
            self._file_chooser_button.set_path(path_to_snd_file)

    def _on_path_picked(self, button: FileChooserButton, file_paths: list[str]) -> None:
        if not file_paths:
            return

        path = file_paths[0]

        model, iter_ = self._ui.sounds_treeview.get_selection().get_selected()
        assert iter_ is not None

        app.settings.set("last_sounds_dir", str(Path(path).parent))

        model[iter_][Column.PATH] = str(path)
        model[iter_][Column.ENABLED] = True

    def _on_clear(self, _button: Gtk.Button) -> None:
        self._file_chooser_button.reset()
        model, iter_ = self._ui.sounds_treeview.get_selection().get_selected()
        assert iter_ is not None

        model[iter_][Column.PATH] = ""
        model[iter_][Column.ENABLED] = False

    def _on_play(self, _button: Gtk.Button) -> None:
        model, iter_ = self._ui.sounds_treeview.get_selection().get_selected()
        assert iter_ is not None

        snd_event_config_name = model[iter_][Column.CONFIG]
        play_sound(snd_event_config_name, None, force=True)

    def _cleanup(self) -> None:
        pass



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Message Retractions (XEP-0424)

from __future__ import annotations

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import MessageReceived
from gajim.common.events import MessageRetracted
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.message import MessageState
from gajim.common.modules.message_util import get_chat_type_and_direction
from gajim.common.modules.message_util import get_message_timestamp
from gajim.common.modules.message_util import get_occupant_info
from gajim.common.modules.message_util import UNKNOWN_MESSAGE
from gajim.common.storage.archive import models as mod
from gajim.common.structs import MessageType
from gajim.common.structs import OutgoingMessage


class Retraction(BaseModule):

    _nbxmpp_extends = 'Retraction'

    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_retraction_message,
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=46,
            ),
        ]

    def send_retraction(
        self,
        contact: types.ChatContactT,
        retract_ids: list[str],
    ) -> None:

        for retraction_id in retract_ids:
            message = OutgoingMessage(
                account=self._account,
                contact=contact,
                retraction_id=retraction_id,
                play_sound=False,
            )

            self._client.send_message(message)
            self._log.info('Send retraction for %s to %s', retraction_id, contact.jid)

    def _process_retraction_message(
        self,
        _client: types.NBXMPPClient,
        stanza: Message,
        properties: MessageProperties,
    ) -> None:

        if properties.retraction is None:
            return

        if properties.moderation is not None:
            # Handled in moderation module
            return

        if properties.type.is_error:
            return

        assert properties.remote_jid is not None
        assert properties.jid is not None

        muc_data = self._client.get_module('MUC').get_muc_data(properties.remote_jid)

        m_type, direction = get_chat_type_and_direction(
            muc_data, self._client.get_own_jid(), properties
        )

        remote_jid = properties.remote_jid
        assert remote_jid is not None

        timestamp = get_message_timestamp(properties)
        occupant = None

        if m_type in (MessageType.GROUPCHAT, MessageType.PM):
            contact = self._client.get_module('Contacts').get_contact(
                properties.jid, groupchat=True
            )
            if not isinstance(contact, GroupchatParticipant):
                self._log.warning(
                    'Ignore unexpected retraction from: %s' % properties.jid
                )
                return

            occupant = get_occupant_info(
                self._account,
                remote_jid,
                self._get_own_bare_jid(),
                direction,
                m_type,
                timestamp,
                contact,
                properties,
            )

        if properties.retraction.is_tombstone:
            assert properties.mam is not None

            self._log.info('Received retraction tombstone from %s for %s',
                           remote_jid, properties.mam.id)

            message_data = mod.Message(
                account_=self._account,
                remote_jid_=remote_jid,
                resource=properties.jid.resource,
                type=m_type,
                direction=direction,
                timestamp=timestamp,
                state=MessageState.ACKNOWLEDGED,
                text=UNKNOWN_MESSAGE,
                id=properties.id,
                stanza_id=properties.mam.id,
                occupant_=occupant,
            )

            pk = app.storage.archive.insert_object(message_data)

            app.ged.raise_event(
                MessageReceived(
                    account=self._account,
                    jid=remote_jid,
                    m_type=m_type,
                    mam=properties.mam,
                    pk=pk,
                )
            )

        else:

            self._log.info('Received retraction from %s for %s',
                           remote_jid, properties.retraction.id)
            retraction = mod.Retraction(
                account_=self._account,
                remote_jid_=remote_jid,
                occupant_=occupant,
                id=properties.retraction.id,
                direction=direction,
                timestamp=properties.retraction.timestamp,
            )

            pk = app.storage.archive.insert_row(retraction, ignore_on_conflict=True)
            if pk == -1:
                raise NodeProcessed

            app.ged.raise_event(
                MessageRetracted(
                    account=self._account, jid=remote_jid, retraction=retraction
                )
            )

        raise NodeProcessed



import sys
from pathlib import Path

from .config import *  # noqa: F403

__version__ = "2.4.5"

portable_path = Path(sys.executable).parent / "is_portable"
IS_PORTABLE = portable_path.exists()

# Check for Microsoft app identity by trying to query AppInfo.
# If there is no AppInfo, it's no MS Store install.
has_ms_appinfo = False
try:
    from winrt.windows.applicationmodel import AppInfo

    has_ms_appinfo = AppInfo.current is not None
except Exception:
    pass

IS_MS_STORE = has_ms_appinfo



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging
from collections.abc import Callable

from gi.repository import Gio

from gajim.common.dbus.base import DBusService
from gajim.common.dbus.menu import DBusMenu
from gajim.common.dbus.menu import DBusMenuService

SNI_NODE_INFO = Gio.DBusNodeInfo.new_for_xml(
    """
<?xml version="1.0" encoding="UTF-8"?>
<node>
    <interface name="org.kde.StatusNotifierItem">
        <property name="Category" type="s" access="read"/>
        <property name="Id" type="s" access="read"/>
        <property name="Title" type="s" access="read"/>
        <property name="ToolTip" type="(sa(iiay)ss)" access="read"/>
        <property name="Menu" type="o" access="read"/>
        <property name="ItemIsMenu" type="b" access="read"/>
        <property name="IconName" type="s" access="read"/>
        <property name="IconThemePath" type="s" access="read"/>
        <property name="Status" type="s" access="read"/>
        <property name="AttentionIconName" type="s" access="read"/>
        <method name="Activate">
            <arg name="x" type="i" direction="in"/>
            <arg name="y" type="i" direction="in"/>
        </method>
        <method name="SecondaryActivate">
            <arg name="x" type="i" direction="in"/>
            <arg name="y" type="i" direction="in"/>
        </method>
        <signal name="NewIcon"/>
        <signal name="NewTooltip"/>
        <signal name="NewStatus">
          <arg name="status" type="s"/>
        </signal>
        <signal name="NewMenu" />
        <signal name="NewTitle" />
        <signal name="NewAttentionIcon" />
        <signal name="NewOverlayIcon" />
        <signal name="NewMenu" />
    </interface>
</node>"""
).interfaces[0]


log = logging.getLogger("gajim.c.tray.linux")


class StatusNotifierItemService(DBusService):
    Category = "Communications"
    Id = "org.gajim.Gajim"
    Title = "Gajim"
    Status = "Active"
    IconName = ""
    IconThemePath = ""
    AttentionIconName = ""
    ToolTip: tuple[str, list[Any], str, str] = ("", [], "Gajim", "")
    ItemIsMenu = False
    Menu = "/StatusNotifierItem/Menu"

    def __init__(
        self,
        session_bus: Gio.DBusConnection,
        menu: DBusMenu,
        theme_path: str | None,
        on_activate_callback: Callable[[], None],
    ):
        super().__init__(
            interface_info=SNI_NODE_INFO,
            object_path="/StatusNotifierItem",
            bus=session_bus,
        )

        if theme_path is not None:
            self.IconThemePath = theme_path

        self._on_activate_callback = on_activate_callback
        self._registered = False
        self._cancellable: Gio.Cancellable | None = None
        self._bus = session_bus
        self._menu = DBusMenuService("/StatusNotifierItem/Menu", session_bus, menu)

    def register(self):
        if self._registered:
            return

        self._registered = True
        self._cancellable = Gio.Cancellable()
        self._menu.register()
        super().register()

        def _on_finish(src: Gio.DBusProxy, res: Gio.AsyncResult) -> None:
            self._cancellable = None
            try:
                proxy = Gio.DBusProxy.new_finish(res)
                proxy.RegisterStatusNotifierItem(  # pyright: ignore
                    "(s)", "/StatusNotifierItem"
                )
            except Exception as error:
                log.error(error)
                return

        Gio.DBusProxy.new(
            connection=self._bus,
            flags=Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
            info=None,
            name="org.kde.StatusNotifierWatcher",
            object_path="/StatusNotifierWatcher",
            interface_name="org.kde.StatusNotifierWatcher",
            cancellable=self._cancellable,
            callback=_on_finish,
        )

    def unregister(self):
        if self._cancellable is not None:
            self._cancellable.cancel()

        super().unregister()
        self._menu.unregister()
        self._registered = False

    def Activate(self, _x: int, _y: int) -> None:
        self._on_activate_callback()

    def SecondaryActivate(self, _x: int, _y: int) -> None:
        self._on_activate_callback()

    def get_visible(self) -> bool:
        return self._registered

    def set_menu(self, menu: DBusMenu) -> None:
        self._menu.set_menu(menu)

    def set_icon(self, icon: str) -> None:
        self.IconName = icon
        self.emit_signal("NewIcon")

    def set_tooltip(self, title: str, description: str) -> None:
        self.ToolTip = ("", [], title, description)
        self.emit_signal("NewTooltip")
